from __future__ import annotations

import logging
import warnings
from html.parser import HTMLParser
from typing import (
    Annotated,
    Any,
    ForwardRef,
    List,
    Literal,
    Type,
    TypedDict,
    TypeVar,
    Union,
)

from pydantic import BaseModel, Field

"""
This is a port of selected parts of the FormKit schema
to Pydantic models.
"""


logger = logging.getLogger(__name__)

HtmlAttrs = dict[str, str | dict[str, str]]


Node = ForwardRef("Node")

# Radio, Select, Autocomplete and Dropdown nodes have
# these options
OptionsType = str | list[dict[str, Any]] | list[str] | dict[str, str] | None


class FormKitSchemaCondition(BaseModel):
    node_type: Literal["condition"] = Field(default="condition", exclude=True)
    if_condition: str = Field(..., alias="if")
    then_condition: Node | List[Node] = Field(..., alias="then")
    else_condition: Node | List[Node] | None = Field(None, alias="else")


class FormKitSchemaMeta(BaseModel):
    __root__: dict[str, str | float | int | bool | None]


class FormKitTypeDefinition(BaseModel): ...


class FormKitContextShape(BaseModel):
    type: Literal["input", "list", "group"]
    value: Any
    _value: Any


class FormKitListValue(BaseModel):
    __root__: str | list[str] | list[dict[str, str]]


class FormKitListStatement(BaseModel):
    """
    A full loop statement in tuple syntax. Can be read like "foreach value, key? in list"
    A 2 or 2 element tuple of value, key, and list or value, list
    """

    __root__: tuple[str, float | int | str, list[FormKitListValue]]


class FormKitSchemaAttributesCondition(BaseModel):
    if_: str = Field(alias="if")
    then_: FormKitAttributeValue = Field(alias="then")
    else_: FormKitAttributeValue | None = Field(alias="else")

    class Config:
        allow_population_by_field_name = True


class FormKitAttributeValue(BaseModel):
    """
    The possible value types of attributes (in the schema)
    """

    __root__: (
        str
        | int
        | float
        | bool
        | None
        | FormKitSchemaAttributes
        | FormKitSchemaAttributesCondition
    )


class FormKitSchemaAttributes(BaseModel):
    __root__: dict[
        str,
        FormKitAttributeValue
        | FormKitSchemaAttributes
        | FormKitSchemaAttributesCondition,
    ]


class FormKitSchemaProps(BaseModel):
    """
    Properties available in all schema nodes.
    """

    # "ForwardRefs" do not work well with django-ninja.
    # This would ideally be:
    # children: str | list[FormKitSchemaProps] | FormKitSchemaCondition | None = Field(
    #     default_factory=list
    # )
    children: str | list[FormKitSchemaProps | str] | FormKitSchemaCondition | None = (
        Field()
    )
    key: str | None
    if_condition: str | None = Field(alias="if")
    for_loop: FormKitListStatement | None = Field(alias="for")
    bind: str | None
    meta: FormKitSchemaMeta | None

    # These are not formal parts of spec, but
    # are attributes defined in ts as Record<string, any>
    # id: str | uuid.UUID | None = Field(None)
    id: str = Field(None)
    name: str | None = Field(None)
    label: str | None = Field(None)
    help: str | None = Field(None)
    validation: str | None = Field(None)
    validationLabel: str | None = Field(None, alias="validation-label")
    validationVisibility: str | None = Field(None, alias="validation-visibility")
    validationMessages: str | dict[str, str] = Field(None, alias="validation-messages")
    placeholder: str | None = Field(None)
    value: str | None = Field(None)
    prefixIcon: str | None = Field(None, alias="prefix-icon")
    classes: str | dict[str, str] | None = Field(None)

    # FormKit allows arbitrary values, we do our best to represent these here
    # Additional Props can be quite a complicated structure
    additional_props: None | dict[str, str | dict[str, Any]] = Field(None)

    class Config:
        allow_population_by_field_name = True

    def dict(self, *args, **kwargs):
        # Set some sensible defaults for "to_dict"
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        if "exclude_none" not in kwargs:
            kwargs["exclude_none"] = True
        _ = super().dict(*args, **kwargs)
        if "additional_props" in _:
            _.update(_["additional_props"])
            del _["additional_props"]
        return _


# We defined this after the model above as it's a circular reference
ChildNodeType = str | list[FormKitSchemaProps | str] | FormKitSchemaCondition | None


class TextNode(FormKitSchemaProps):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["text"] = Field(default="text", alias="$formkit")
    text: str | None


class TextAreaNode(TextNode):
    formkit: Literal["textarea"] = Field(default="textarea", alias="$formkit")
    text: str | None


class DateNode(TextNode):
    formkit: Literal["date"] = Field(default="date", alias="$formkit")


class CurrencyNode(TextNode):
    formkit: Literal["currency"] = Field(default="currency", alias="$formkit")


class UuidNode(TextNode):
    formkit: Literal["uuid"] = Field(default="uuid", alias="$formkit")


class DatePickerNode(TextNode):
    formkit: Literal["datepicker"] = Field(default="datepicker", alias="$formkit")
    calendarIcon: str = "calendar"
    format: str = "DD/MM/YY"
    nextIcon: str = "angleRight"
    prevIcon: str = "angleLeft"


class CheckBoxNode(TextNode):
    formkit: Literal["checkbox"] = Field(default="checkbox", alias="$formkit")


class NumberNode(TextNode):
    formkit: Literal["number"] = Field(default="number", alias="$formkit")
    text: str | None
    max: int | None = None
    min: int | None = None
    step: str | None = None


class PasswordNode(TextNode):
    formkit: Literal["password"] = Field(default="password", alias="$formkit")
    name: str | None


class HiddenNode(TextNode):
    formkit: Literal["hidden"] = Field(default="hidden", alias="$formkit")


class RadioNode(TextNode):
    formkit: Literal["radio"] = Field(default="radio", alias="$formkit")
    name: str | None
    options: OptionsType = Field(None)


class SelectNode(TextNode):
    formkit: Literal["select"] = Field(default="select", alias="$formkit")
    options: OptionsType = Field(None)


class AutocompleteNode(TextNode):
    formkit: Literal["autocomplete"] = Field(default="autocomplete", alias="$formkit")
    options: OptionsType = Field(None)


class EmailNode(TextNode):
    formkit: Literal["email"] = Field(default="email", alias="$formkit")


class TelNode(TextNode):
    formkit: Literal["tel"] = Field(default="tel", alias="$formkit")


class DropDownNode(TextNode):
    formkit: Literal["dropdown"] = Field(default="dropdown", alias="$formkit")
    options: OptionsType = Field(None)
    empty_message: str | None = Field(None, alias="empty-message")
    select_icon: str | None = Field(None, alias="selectIcon")
    placeholder: str | None


class RepeaterNode(TextNode):
    formkit: Literal["repeater"] = Field(default="repeater", alias="$formkit")
    up_control: bool | None = Field(default=True, alias="upControl")
    down_control: bool | None = Field(default=True, alias="downControl")
    add_label: str | None = Field(default="Add another", alias="addLabel")
    name: str | None = None


class GroupNode(TextNode):
    formkit: Literal["group"] = Field(default="group", alias="$formkit")
    text: str | None


# This is useful for "isinstance" checks
# which do not work with "Annotated" below
FormKitType = (
    TextNode
    | TextAreaNode
    | CheckBoxNode
    | PasswordNode
    | SelectNode
    | AutocompleteNode
    | EmailNode
    | NumberNode
    | RadioNode
    | GroupNode
    | DateNode
    | DatePickerNode
    | DropDownNode
    | RepeaterNode
    | TelNode
    | CurrencyNode
    | HiddenNode
    | UuidNode
)

FormKitSchemaFormKit = Annotated[
    Union[
        TextNode,
        TextAreaNode,
        CheckBoxNode,
        PasswordNode,
        SelectNode,
        AutocompleteNode,
        EmailNode,
        NumberNode,
        RadioNode,
        GroupNode,
        DateNode,
        DatePickerNode,
        DropDownNode,
        RepeaterNode,
        TelNode,
        CurrencyNode,
        HiddenNode,
        UuidNode,
    ],
    Field(discriminator="formkit"),
]


class FormKitSchemaDOMNode(FormKitSchemaProps):
    """
    HTML elements are defined using the $el property.
    You can use $el to render any HTML element.
    Attributes can be added with the attrs property,
    and content is assigned with the children property
    """

    node_type: Literal["element"] = Field(default="element", exclude=True)
    el: str = Field(alias="$el")
    attrs: FormKitSchemaAttributes | None

    class Config:
        allow_population_by_field_name = True


class FormKitSchemaComponent(FormKitSchemaProps):
    """
    Components can be defined with the $cmp property
    The $cmp property should be a string that references
    a globally defined component or a component passed
    into FormKitSchema with the library prop.
    """

    node_type: Literal["component"] = Field(default="component", exclude=True)

    cmp: str = Field(
        ...,
        alias="$cmp",
        description="The $cmp property should be a string that references a globally defined component or a component passed into FormKitSchema with the library prop.",  # noqa: E501
    )
    props: dict[str, str | Any] | None

    class Config:
        allow_population_by_field_name = True


# This necessary to properly "populate" some more complicated models
FormKitSchemaAttributesCondition.update_forward_refs()
FormKitAttributeValue.update_forward_refs()
FormKitSchemaDOMNode.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
FormKitSchemaComponent.update_forward_refs()


class FormKitTagParser(HTMLParser):
    """
    Reverse an HTML example to schema
    This is for lazy copy-pasting from the formkit website :)
    """

    def __init__(self, html_content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data: str | None = None

        self.current_tag: FormKitSchemaFormKit | None = None
        self.tags: list[FormKitSchemaFormKit] = []
        self.parents: list[FormKitSchemaFormKit] = []
        self.feed(html_content)

    def handle_starttag(self, tag, attrs):
        """
        Read anything that's a "formtag" type
        """
        if tag != "formkit":
            return
        props = dict(attrs)
        props["formkit"] = props.pop("type")

        tag = FormKitSchemaFormKit(**props)
        self.current_tag = tag

        if self.parents:
            self.parents[-1].children.append(tag)
        else:
            self.tags.append(tag)
            self.parents.append(tag)

    def handle_endtag(self, tag: str) -> None:
        if tag != "formkit":
            return
        if self.parents:
            self.parents.pop()

    def handle_data(self, data):
        if self.current_tag and data.strip():
            self.current_tag.children.append(data.strip())
            # Ensure that children is included even when "exclude_unset" is True
            # since we populated this after the initial tag build
            self.current_tag.__fields_set__.add("children")


FormKitSchemaDOMNode.update_forward_refs()

Model = TypeVar("Model", bound="BaseModel")
StrBytes = str | bytes

Node = Annotated[
    Union[
        FormKitSchemaFormKit,
        FormKitSchemaDOMNode,
        FormKitSchemaComponent,
        FormKitSchemaCondition,
    ],
    Field(discriminator="node_type"),
]

NODE_TYPE = Literal["condition", "formkit", "element", "component"]
FORMKIT_TYPE = Literal[
    "text",
    "textarea",
    "tel",
    "currency",
    "select",
    "checkbox",
    "number",
    "group",
    "list",
    "password",
    "button",
    "radio",
    "form",
    "date",
    "datepicker",
    "dropdown",
    "repeater",
    "autocomplete",
    "email",
    "uuid",
]


class Discriminators(TypedDict, total=False):
    node_type: NODE_TYPE
    formkit: FORMKIT_TYPE


def get_node_type(obj: dict) -> Discriminators:
    """
    Pydantic requires nodes to be "differentiated" by a field value
    when used in a Union type situation.
    This function should return the 'node_type' values and if present 'Formkit' value
    which corresponds to the object being inspected.
    """
    if "__root__" in obj:
        return get_node_type(obj["__root__"])

    if isinstance(obj, str):
        return "text"

    if isinstance(obj, dict) and len(obj.keys()) == 0:
        return "text"

    for key, return_value in (
        ("$el", "element"),
        ("$formkit", "formkit"),
        ("$cmp", "component"),
    ):
        if key in obj:
            return return_value
    raise KeyError(f"Could not determine node type for {obj}")


NodeTypes = (
    FormKitType | FormKitSchemaDOMNode | FormKitSchemaComponent | FormKitSchemaCondition
)


class FormKitNode(BaseModel):
    __root__: str | Node

    @classmethod
    def parse_obj(
        cls: Type["Model"], obj: str | dict, recursive: bool = True
    ) -> "Model":
        """
        This classmethod differentiates between the different "Node" types
        when deserializing
        """

        def get_additional_props(object_in: dict[str, Any], exclude: set[str] = set()):
            """
            Parse the object or database return (dict)
            to break out fields we handle in JSON

            A FormKit node can have 'arbitrary' additional properties
            For instance classes to apply to child nodes
            here we can't realistically cover every scenario so
            fall back to JSON storage for thes

            However: if we're coming from the database we already store these in a separate field
            """
            # Things which are not "other attributes"
            set_handled_keys = {
                "$formkit",
                "$el",
                "if",
                "for",
                "then",
                "else",
                "children",
                "node_type",
                "formkit",
                "id",
            }
            # Merge "additional props" from the input object
            # with any "unknown" params we received
            props: dict[str, Any] = object_in.get("additional_props", {})
            props.update(
                {k: obj[k] for k in object_in.keys() - exclude - set_handled_keys}
            )
            return props

        def get_children(object_in: dict):
            if children_in := object_in.get("children", None):
                if isinstance(children_in, str):
                    children_in = [children_in]

                children_out = []
                for n in children_in:
                    if isinstance(n, str):
                        children_out.append(n)
                    else:
                        try:
                            children_out.append(cls.parse_obj(n).__root__)
                        except Exception as E:
                            warnings.warn(f"{E}")
                return children_out
            else:
                return None

        if isinstance(obj, str):
            return obj

        # There's a discriminator step which needs assisance: `node_type`
        # must be set on the input object
        try:
            node_type = get_node_type(obj)
        except Exception as E:
            raise KeyError(f"Node type couln't be determined: {obj}") from E

        try:
            parsed = super().parse_obj({**obj, "node_type": node_type})
            node: NodeTypes = parsed.__root__
        except KeyError as E:
            raise KeyError(f"Unable to parse content {obj} to a {cls}") from E
        if additional_props := get_additional_props(obj, exclude=set(node.__fields__)):
            node.additional_props = additional_props
        # Recursively parse 'child' nodes back to Pydantic models for 'children'
        if recursive:
            node.children = get_children(obj)
        else:
            node.children = None
        return parsed


class FormKitSchema(BaseModel):
    __root__: list[Node]

    @classmethod
    def parse_obj(cls: Type["Model"], obj: Any) -> "Model":
        """
        Parse a set of FormKit nodes or a single 'GroupNode' to
        a 'schema'
        """
        # If we're parsing a single node, wrap it in a list
        if isinstance(obj, dict):
            return cls.parse_obj([obj])
        try:
            return cls(__root__=[FormKitNode.parse_obj(_).__root__ for _ in obj])
        except TypeError:
            raise


FormKitSchema.update_forward_refs()
FormKitSchemaCondition.update_forward_refs()
PasswordNode.update_forward_refs()

FormKitSchemaDefinition = Node | list[Node] | FormKitSchemaCondition



# ruff: noqa: F401 F811
# flake8: noqa: F401 F811

import os
from importlib.util import find_spec
from typing import Type

import pytest

from formkit_ninja import models

os.environ.setdefault("DJANGO_ALLOW_ASYNC_UNSAFE", "true")

playwright = pytest.mark.skipif(
    # 'playwright' is not installed
    find_spec("playwright") is None or find_spec("pytest_playwright") is None,
    reason="playwright is not installed",
)


@playwright
def test_home_page(admin_page):
    admin_page.get_by_role("link", name="Form kit schema nodes").click()
    admin_page.get_by_role("link", name="Add form kit schema node").click()
    # There was a bug identified by this test where "label" in JSON
    # would override "label" in the parent model
    admin_page.get_by_label("Label:").fill("test")
    admin_page.get_by_label("Node type:").select_option("$formkit")
    admin_page.get_by_label("Description:").fill("This is a test node")
    admin_page.get_by_role("button", name="Save and continue editing").click()
    admin_page.get_by_label("Name:").fill("test")
    admin_page.get_by_label("Key:").fill("test")
    admin_page.get_by_label("Placeholder:").fill("Please select something....")
    admin_page.get_by_label("Html id:").fill("test")
    admin_page.get_by_role("button", name="Save", exact=True).click()
    admin_page.get_by_role("link", name="Form kit schemas").click()
    admin_page.get_by_role("link", name="Add form kit schema").click()
    admin_page.get_by_label("Label:").fill("Test Schema")
    admin_page.get_by_role("button", name="Save and continue editing").click()
    admin_page.get_by_role("link", name="Add another Form components").click()
    admin_page.get_by_role("button", name="Save", exact=True).click()
    admin_page.get_by_role("link", name="Form componentss").click()
    admin_page.get_by_role("link", name="Add form components").click()
    admin_page.get_by_label("Schema:").select_option(index=0)
    admin_page.get_by_label("Node:").select_option(index=0)
    admin_page.get_by_label("Label:").fill("'test' node in 'test' schema")
    admin_page.get_by_role("button", name="Save", exact=True).click()


# Tests that the admin pages from importing SF 1.1 make sense


@pytest.mark.django_db()
def test_import_sf11(SF_1_1):
    """
    This tests that we can successfully import the 'SF11' form from Partisipa
    """
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema = BaseModel.parse_obj(SF_1_1)
    schema_in_the_db = models.FormKitSchema.from_pydantic(schema)

    # Check that schema sections retained their order
    # There is a top level node
    assert schema_in_the_db.nodes.first().label == "SF_1_1"

    # The 'top level' node has the same child nodes as the input schema
    child_nodes = schema_in_the_db.nodes.first().get_node(recursive=True).children

    assert [n["id"] for n in SF_1_1["children"]] == [n.id for n in child_nodes]

    # The inputs and outputs
    partA_in = SF_1_1["children"][0]["children"]
    partA_schema = child_nodes[0].children
    partA_out = [n.dict() for n in partA_schema]

    # Nested (children) should retain their order
    assert (
        [n["key"] for n in partA_in]
        == [n.key for n in partA_schema]
        == [n["key"] for n in partA_out]
    )
    assert (
        [n["label"] for n in partA_in]
        == [n.label for n in partA_schema]
        == [n["label"] for n in partA_out]
    )

    assert partA_in[0]["key"] == partA_schema[0].key
    assert partA_in[0]["name"] == partA_schema[0].name
    assert partA_in[0]["label"] == partA_schema[0].label
    assert partA_in[0]["placeholder"] == partA_schema[0].placeholder
    assert partA_in[0]["options"] == partA_schema[0].options

    assert partA_in[0]["key"] == partA_out[0]["key"]
    assert partA_in[0]["name"] == partA_out[0]["name"]
    assert partA_in[0]["label"] == partA_out[0]["label"]
    assert partA_in[0]["placeholder"] == partA_out[0]["placeholder"]
    assert partA_in[0]["options"] == partA_out[0]["options"]


@playwright
@pytest.mark.django_db()
def test_admin_actions_sf11(SF_1_1, admin_page):
    """
    This tests that we can successfully import the 'SF11' form from Partisipa
    """
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema = BaseModel.parse_obj(SF_1_1)
    models.FormKitSchema.from_pydantic(schema)
    admin_page.get_by_role("link", name="Form kit schema nodes").click()
    ...


@playwright
@pytest.mark.django_db()
def test_import_1321(TF_13_2_1, admin_page):
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    models.FormKitSchema.from_pydantic(BaseModel.parse_obj(TF_13_2_1))
    admin_page.get_by_role("link", name="Form kit schema nodes").click()
    admin_page.get_by_role("link", name="repeaterProjectProgress").click()
    admin_page.pause()
    ...


@playwright
@pytest.mark.django_db()
def test_admin_all_forms(
    CFM_12_FF_12,
    CFM_2_FF_4,
    FF_14,
    POM_1,
    SF_1_1,
    SF_1_2,
    SF_1_3,
    SF_2_3,
    SF_4_1,
    SF_4_2,
    SF_6_2,
    TF_13_2_1,
    TF_6_1_1,
    admin_page,
):
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    _ = list(
        (
            models.FormKitSchema.from_pydantic(BaseModel.parse_obj(n))
            for n in (
                CFM_12_FF_12,
                CFM_2_FF_4,
                FF_14,
                POM_1,
                SF_1_1,
                SF_1_2,
                SF_1_3,
                SF_2_3,
                SF_4_1,
                SF_4_2,
                SF_6_2,
                TF_13_2_1,
                TF_6_1_1,
            )
        )
    )

    admin_page.pause()



import json
from importlib.resources import files

from formkit_ninja import formkit_schema, samples


def test_node_parse():
    schema = json.loads(files(samples).joinpath("element.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])
    formkit_schema.FormKitSchema.parse_obj(schema)


def test_login_form():
    schema = json.loads(
        files(samples).joinpath("form_generation_example.json").read_text()
    )
    formkit_schema.FormKitNode.parse_obj(schema[0])
    formkit_schema.FormKitSchema.parse_obj(schema)


def test_raw_values():
    schema = json.loads(files(samples).joinpath("raw_values.json").read_text())
    node = formkit_schema.FormKitNode.parse_obj(schema)
    node_reloaded = json.loads(node.json(by_alias=True, exclude_unset=True))
    assert node_reloaded["props"]["__raw__price"] == "$2.99"
    assert node_reloaded["$cmp"] == "PriceComponent"


def test_meeting_type_node():
    """
    Loads a single element schema, checking that return values
    are the same as entered values
    """
    schema = json.loads(files(samples).joinpath("meeting_type_node.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])
    meeting_type_schema = formkit_schema.FormKitSchema.parse_obj(schema)
    reloaded = json.loads(meeting_type_schema.json(by_alias=True, exclude_none=True))

    assert reloaded[0].get("id") == "meeting_type"
    assert reloaded[0].get("name") == "meeting_type"
    assert reloaded[0].get("$formkit") == "select"

    assert len(reloaded[0]["options"]) == 2


def test_repeater():
    schema = json.loads(files(samples).joinpath("repeater.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])


def test_dropdown():
    schema = json.loads(files(samples).joinpath("dropdown.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])



import json
from importlib.resources import files

import pytest

from formkit_ninja import formkit_schema, models, samples
from formkit_ninja.formkit_schema import FormKitNode, FormKitSchemaDOMNode
from formkit_ninja.schemas import Schemas

schemas = Schemas()


def _read_file(filename: str):
    return json.loads(files(samples).joinpath(f"{filename}.json").read_text())


def _parse_file(filename: str):
    return formkit_schema.FormKitSchema.parse_obj(_read_file(filename))


@pytest.fixture(params=list(schemas))
def schema(request):
    return s.as_json(request.param)


def test_node_parse():
    schema = json.loads(files(samples).joinpath("element.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])
    _parse_file("element")


@pytest.mark.django_db()
def test_create_schema():
    c = models.FormKitSchema.objects.create()  # noqa: F841


@pytest.mark.django_db()
def test_create_from_schema():
    """
    Create a database entry from a FormKit schema
    """
    element_schema = _parse_file("element")
    c = models.FormKitSchema.from_pydantic(element_schema)
    # This test schema has one node
    assert c.nodes.count() == 1

    # The node...
    dictify = c.nodes.first().get_node(recursive=True).dict(exclude_none=True)

    # The node looks like this:
    # {'children': [], 'node_type': 'element', 'el': 'div', 'attrs': {'style': {...}, 'data-foo': 'bar'}}

    assert set(dictify.keys()) == {"children", "$el", "attrs"}
    assert dictify["children"] == ["Hello world"]
    assert dictify["$el"] == "div"
    assert set(dictify["attrs"].keys()) == {"style", "data-foo"}

    # A 'group' node
    first_node = c.nodes.first()
    first_node.get_node().dict(exclude_none=True, exclude={"children"})


@pytest.mark.django_db()
def test_parse_el_priority(el_priority: dict):
    """
    Regression test for missing text content

    Expects: {
        "$el": "div",
        "children": ["Priority ", {"$el": "span", "attrs": {"class": "ml-1"}, "children": ["$: ($index + 1)"]}],
        "attrs": {"class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"},
    }
    """
    parsed_node = FormKitNode.parse_obj(el_priority).__root__
    assert isinstance(parsed_node, FormKitSchemaDOMNode)
    # Load it into the database
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    # The node type in the database is `$el`
    assert node_in_the_db.node_type == "$el"

    el_priority["children"][0] == "Priority "
    children: list[models.FormKitSchemaNode] = list(
        node_in_the_db.children.all().order_by("nodechildren__order")
    )

    # The first child is an 'text'
    node = children[0]
    assert node.node_type == "text"
    assert node.text_content == "Priority "
    assert node.to_pydantic() == "Priority "

    # The second child is an 'el'
    node = children[1]
    assert node.node_type == "$el"
    assert node.node.get("$el") == "span"
    assert node.node["attrs"] == {"class": "ml-1"}
    # With appropriate `by_alias` and `exclude_defaults` we should get an equal output as input
    assert (
        node.to_pydantic(recursive=True).dict(by_alias=True, exclude_defaults=True)[
            "__root__"
        ]
        == el_priority["children"][1]
    )


@pytest.mark.django_db()
def test_parse_simple_text_node(simple_text_node: dict):
    node: FormKitNode = FormKitNode.parse_obj(simple_text_node)
    parsed_node: FormKitSchemaDOMNode = node.__root__
    # Before the fix, this was breaking into individual letters
    assert parsed_node.children == ["Priority"]


@pytest.mark.django_db()
def test_additional_props(formkit_text_node: dict):
    """
    This ensures that custom attributes are maintained
    from JSON to Pydantic,
    from Pydantic to the Database,
    from the Database to Pydantic,
    and from Pydantic to JSON.
    """
    node: FormKitNode = FormKitNode.parse_obj(formkit_text_node)

    # From JSON to a Pydantic class...
    parsed_node: formkit_schema.SelectNode = node.__root__
    assert parsed_node.additional_props == {"class": "red"}

    # Into the database...
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    assert node_in_the_db.additional_props == {"class": "red"}
    node_in_the_db.to_pydantic()

    # The 'discriminator' fields should not be stored in the db
    assert node_in_the_db.node["$formkit"] == "select"
    assert set(node_in_the_db.node.keys()) == {
        "key",
        "id",
        "name",
        "label",
        "placeholder",
        "$formkit",
    }
    # And out of the database again
    from_the_db = node_in_the_db.to_pydantic(options=True)
    assert from_the_db.__root__.additional_props == {"class": "red"}

    # And back to JSON
    json_from_the_db = json.loads(
        from_the_db.json(exclude_none=True, by_alias=True, exclude={"node_type"})
    )
    assert json_from_the_db["class"] == "red"

    # Additional checks that the JSON output is equivalent to the JSON input
    # Note that json from the db has additional Python 'discriminator' fields 'node_type' and 'formkit'

    assert set(node_in_the_db.node.keys()) == {
        "key",
        "id",
        "name",
        "label",
        "placeholder",
        "$formkit",
    }

    assert node_in_the_db.node.get("key") == formkit_text_node["key"]
    assert node_in_the_db.node.get("id") == formkit_text_node["id"]
    assert node_in_the_db.node.get("name") == formkit_text_node["name"]
    assert node_in_the_db.node.get("label") == formkit_text_node["label"]
    assert node_in_the_db.node.get("$formkit") == formkit_text_node["$formkit"]
    assert node_in_the_db.node.get("placeholder") == formkit_text_node["placeholder"]
    assert node_in_the_db.additional_props == {"class": "red"}

    assert json_from_the_db["key"] == formkit_text_node["key"]
    assert json_from_the_db["id"] == formkit_text_node["id"]
    assert json_from_the_db["name"] == formkit_text_node["name"]
    assert json_from_the_db["label"] == formkit_text_node["label"]
    assert json_from_the_db["$formkit"] == formkit_text_node["$formkit"]
    assert json_from_the_db["placeholder"] == formkit_text_node["placeholder"]
    # This "additional prop" is now merged again to the json data
    assert json_from_the_db["class"] == formkit_text_node["class"]
    assert json_from_the_db["options"] == formkit_text_node["options"]


s = Schemas()
schemas = s.list_schemas()


def schema_are_same(in_: dict | str, out_: dict | str):
    if isinstance(in_, str):
        assert in_ == out_, f"{in_=} != {out_=}"
        return

    assert set(in_.keys()) == set(out_.keys())
    for key in in_.keys():
        if key in {"options", "children"}:
            continue
        assert in_[key] == out_[key], f"{in_[key]=} != {out_[key]=}"
    if "children" in in_:
        for c_in, c_out in zip(in_.get("children", []), out_.get("children", [])):
            schema_are_same(c_in, c_out)


@pytest.mark.django_db()
def test_schemas(schema: dict):
    node: FormKitNode = FormKitNode.parse_obj(schema, recursive=True)
    parsed_node: formkit_schema.SelectNode = node.__root__
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]

    # Returning the code
    schema_out: dict = node_in_the_db.to_pydantic(recursive=True, options=True).dict(
        by_alias=True, exclude_none=True
    )["__root__"]
    schema_are_same(schema, schema_out)


@pytest.mark.django_db()
def test_protected_model(formkit_text_node: dict):
    """
    A 'protected' node cannot be deleted
    This works right at the database level, not through the API
    We can test directly with a call to `.delete`
    """
    node: FormKitNode = FormKitNode.parse_obj(formkit_text_node)
    parsed_node: formkit_schema.SelectNode = node.__root__
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    node_in_the_db.protected = True
    node_in_the_db.save()
    from django.db.utils import InternalError

    with pytest.raises(InternalError):
        node_in_the_db.delete()
